###Imports
import slimPre
import numpy as np
import ogr
import calendar
from scipy.spatial import KDTree
from netCDF4 import Dataset
from scipy.interpolate import RegularGridInterpolator, interp1d, interp2d
import scipy.spatial.qhull as qhull

from os import listdir
from os.path import isfile, join

from datetime import datetime

#%%

###File paths etc.
years = '2012_2013'
pathin = '/home/2/jc231398/TorresStrait/'

MeshFiPath = pathin + 'Mesh/torresV2.msh'
WindBase   = pathin + 'Prepro/Data/Wind/BARRA_R/' +years +'/'
BranBase   = pathin + 'Prepro/Data/OpenBoundaries/BRAN_2016/' +years +'/'
GridSpecFiPath = pathin + 'Prepro/Data/OpenBoundaries/BRAN_2016/grid_spec.nc'
BathFiPath = pathin + 'Prepro/Data/layers/ausbath_09_v4.tif'

ManningSeaGrass = pathin + "Prepro/Data/layers/Original_Seagrass_Polygon_UTM54s.shp"
ManningReefs    = pathin + "Prepro/Data/layers/ReefsDry.shp"

data_dir_base = pathin + 'Prepro/' +years +'_' +slimPre.partition_nb() +'/'
data_dir = data_dir_base +slimPre.partition_id() +'/'
slimPre.make_directory(data_dir)

#Min Depth
minDepth = 200  #only take UV values above this level (200m)

#Topex time step
TStep = 1800

###Set up region
#nb_part = 4
#slimPre.dgpy.dgMeshPartition(MeshFiPath,nb_part)
#mesh_file_name = pathin + "Mesh/torres_%i.msh"%nb_part
mesh_file_name =  slimPre.partition_mesh(MeshFiPath, data_dir+"mesh.msh")

mesh = slimPre.Mesh(mesh_file_name, mesh_proj='+proj=utm +ellps=WGS84 +zone=54 +south')
region_global = slimPre.Region(mesh)
region_global_lonlat = slimPre.Coordinate_system(region_global,"+proj=longlat +ellps=WGS84")
lonlat_points = region_global_lonlat.coordinates
dist = slimPre.get_distance(region_global,"coast",100)

#%%

###Functions
def get_index_of_points_in_polygons(points,polygonshp) :
    ds = ogr.Open(polygonshp)
    lyr_in = ds.GetLayer(0)
    point_ref=ogr.osr.SpatialReference()
    point_ref.ImportFromProj4("+proj=latlong +ellps=WGW84")
    print(lyr_in.GetSpatialRef().ExportToProj4())
    ctran=ogr.osr.CoordinateTransformation(point_ref,lyr_in.GetSpatialRef())
    def check(x,y):
        [lon,lat,z]=ctran.TransformPoint(x,y)
        pt = ogr.Geometry(ogr.wkbPoint)
        pt.SetPoint_2D(0, lon,lat)
        lyr_in.SetSpatialFilter(pt)
        return len(lyr_in) > 0

    mask = [i for i in range(points.shape[0]) if check(points[i,0],points[i,1])]
    return mask

def interp_weights(xyz, uvw):
    tri = qhull.Delaunay(xyz)
    simplex = tri.find_simplex(uvw)
    out = np.where(simplex == -1)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uvw - temp[:, 2]
    bary = np.einsum('njk,nk->nj', temp[:, :2, :], delta)
    weights = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))
    if out :
        tree = KDTree(xyz)
        dout,vertices[out] = tree.query(uvw[out],3)
        wout = 1/dout
        weights[out] = wout / np.sum(wout,axis=1,keepdims=True)
    return vertices, weights

def interpolate(values, vtx, wts):
    return np.einsum('nj,nj->n', np.take(values, vtx), wts)

class WindShadow :
    def __init__(self,mesh,tags,points,distance):
        self.region = slimPre.Region(mesh,tags)
        self.edges = self.region.coordinates[self.region.bnd_elements,:2]
        ec = np.mean(self.edges,axis=1)
        #add 2% tol
        self.edges += (self.edges-ec[:,None,:])*0.02
        treecoast = KDTree(ec)
        treedata = KDTree(points[:,:2])
        p = treecoast.query_ball_tree(treedata,distance)
        self.candidates = np.array([(n,i) for n,pp in enumerate(p) for i in pp])
        self.datax = points
        self.d = distance

    def _get_mask(self,uwind,vwind) :
        w = np.hstack((uwind[self.candidates[:,1],None],vwind[self.candidates[:,1],None]))
        nw = np.hypot(w[:,0],w[:,1])
        ecoast = self.edges[self.candidates[:,0],:,:]
        edata = np.zeros((self.candidates.shape[0],2,2))
        edata[:,:,:] = self.datax[self.candidates[:,1],None,:2]
        edata[:,1,:] -= w * self.d/nw[:,None]
        ec = np.mean(edata,axis=1)
        #add 2% tol
        edata += (edata-ec[:,None,:])*0.02
        def ccw(A,B,C):
            return (C[:,1]-A[:,1])*(B[:,0]-A[:,0]) >= (B[:,1]-A[:,1])*(C[:,0]-A[:,0])
        def intersect(A,B,C,D):
            return np.logical_and(ccw(A,C,D) != ccw(B,C,D), ccw(A,B,C) != ccw(A,B,D))
        def cross(A,B,C,D):
            AB = B-A
            CD = D-C
            return AB[:,0]*CD[:,1]-AB[:,1]*CD[:,0]
        mask = np.logical_and(
                intersect(edata[:,0,:],edata[:,1,:],ecoast[:,0,:],ecoast[:,1,:]),
                cross(edata[:,0,:],edata[:,1,:],ecoast[:,0,:],ecoast[:,1,:]) > 0)
        return mask

    def apply(self,uwind,vwind,f=0) :
        mask = self._get_mask(uwind,vwind)
        uwind[self.candidates[mask,1]] = uwind[self.candidates[mask,1]]*f
        vwind[self.candidates[mask,1]] = vwind[self.candidates[mask,1]]*f

    def set_value(self,uwind,vwind,data,v) :
        mask = self._get_mask(uwind,vwind)
        data[self.candidates[mask,1]] = v

#%%
### Base ###
def prepro_base() :
   
    ###Bathymetry
    bath = -slimPre.geo_tiff(BathFiPath).evaluate_region(region_global)
    bath[bath > 200] = 200

    #Add a deep area in the west so that the model doesn't break on that one reef
    west = slimPre.Region(mesh, 'west')
    d = slimPre.distance_msh(west, 100).evaluate_region(region_global)
    dfactor = 1-np.minimum(d/20000,1)
    bath = np.maximum(bath, 3+7*dfactor)

    slimPre.write_file(data_dir + 'bathymetry.nc', region=region_global, time=None, data=[('bathymetry', bath)])
    slimPre.smooth_bathymetry(mesh, (data_dir+'bathymetry.nc','bathymetry'), output_file_name=data_dir_base+'bathymetry_smooth', coefficient=0.3)    

    #Coriolis
    corio = 2 * 7.292e-5 * np.sin(lonlat_points[:,1])
    slimPre.write_file(data_dir + 'coriolis.nc', region=region_global, time=None, data=[('coriolis', corio)])

    #Manning coefficient
    points = lonlat_points*180/np.pi
    Mdata = np.zeros([points.shape[0]])
    Mdata[:] = 0.023
    Mdata[get_index_of_points_in_polygons(points, ManningSeaGrass)] = 0.055
    Mdata[get_index_of_points_in_polygons(points,ManningReefs)] = 0.072
    slimPre.write_file(data_dir + "manning.nc", region=region_global, data=[("manning",Mdata)])

#%%
### Wind ###
def prepro_wind():
    print('Preprocessing wind data')
    ###Coordmesh
    coordmesh = lonlat_points[:,:2] * 180./np.pi

    ###File lists
    Uwind_files_list = [f for f in listdir(WindBase + 'UWind10m') if isfile(join(WindBase + 'UWind10m', f))]
    Vwind_files_list = [f for f in listdir(WindBase + 'VWind10m') if isfile(join(WindBase + 'VWind10m', f))]

    Uwind_files_list.sort()
    Vwind_files_list.sort()

    ###Array creation
    t = np.zeros(len(Vwind_files_list) )
    uwind = np.zeros((len(t),coordmesh.shape[0]))
    vwind = np.zeros((len(t),coordmesh.shape[0]))
    beaching = np.zeros((len(t),coordmesh.shape[0]))
    wshadow = WindShadow(mesh,"coast",region_global.coordinates,2000)
    wshadowBeaching = WindShadow(mesh,"coast",region_global.coordinates,1000)


    ###Loop for Interpolation to SLIM mesh
    for i in range( len(Vwind_files_list)):
        ###Import U and V files
        u_file = Dataset(WindBase + 'UWind10m/' + Uwind_files_list[i], 'r')
        v_file = Dataset(WindBase + 'VWind10m/' + Vwind_files_list[i], 'r')

        ###Time   
        t[i] = u_file.variables['time'][:]*3600+calendar.timegm((1970,1,1,0,0,0))
 
        ###UV
        udata = np.transpose(u_file.variables['uwnd10m'])
        vdata = np.transpose(v_file.variables['vwnd10m'])
    
        ###Latitude/Longitude
        ulondata = u_file.variables['longitude'][:]
        ulatdata = u_file.variables['latitude'][:]

        vlondata = v_file.variables['longitude'][:]
        vlatdata = v_file.variables['latitude'][:]


        ###Interpolation
        ull = RegularGridInterpolator((ulondata,ulatdata),udata,method='linear')(coordmesh)
        vll = RegularGridInterpolator((vlondata,vlatdata),vdata,method='linear')(coordmesh)
        uwind[i,:], vwind[i,:] = region_global_lonlat.rotate(ull,vll)
        wshadowBeaching.set_value(-uwind[i,:],-vwind[i,:],beaching[i,:],1)
        wshadow.apply(uwind[i,:],vwind[i,:],0)
        ###End of loop

    ###Slim time
    tslim = slimPre.Time(time_vector=t)

    ###Write...
    slimPre.write_file(data_dir + 'wind.nc', region_global, tslim, 
                   [('u',uwind),('v',vwind),('beaching',beaching)])

#%%
### Open Sea ###
def prepro_open():
    print('Preprocessing Open Sea Boundary')
    
    ###File lists
    UBran_files_list = [f for f in listdir(BranBase + 'u') if isfile(join(BranBase + 'u', f))]
    VBran_files_list = [f for f in listdir(BranBase + 'v') if isfile(join(BranBase + 'v', f))]
    EBran_files_list = [f for f in listdir(BranBase + 'eta') if isfile(join(BranBase + 'eta', f))]
    
    UBran_files_list.sort()
    VBran_files_list.sort()
    EBran_files_list.sort()    
    
    ###Pre calculate no. time iterations. Make mask.
    region_open = slimPre.Region(mesh, ['west', 'south_west', 'coral_sea', 'south_east', 'north_east'])
    lonlat_sys = slimPre.Coordinate_system(region_open, data_proj='+proj=latlong +ellps=WGS84')
    coordmesh = lonlat_sys.coordinates[:,:2] * 180./np.pi
    
    LpIter = 0
    for i in range( len(EBran_files_list)):
        ###Import e file
        e_file = Dataset(BranBase + 'eta/' + EBran_files_list[i], 'r')
        LpIter += e_file.variables['Time'].shape[0]
        
        if i == 0: ###Make mask    
            #u-cell mask
            u_file = Dataset(BranBase + 'u/' + UBran_files_list[i], 'r')

            uc_londata = np.array(u_file.variables['xu_ocean'][:])
            uc_latdata = np.array(u_file.variables['yu_ocean'][:])
            uc_londata, uc_latdata = np.meshgrid(uc_londata,uc_latdata)
            
            u0data = u_file.variables['u'][0,0,:,:]
            uc_mask = ~u0data.mask  #~ bit wise negation
            
            uc_londata = uc_londata[uc_mask]
            uc_latdata = uc_latdata[uc_mask]
            uc_coorddata = np.hstack((uc_londata[:,None], uc_latdata[:,None]))


            #Vertical weights. Part 1.        
            tc_depth = np.array(u_file.variables['st_ocean'][:])
            zcSel = tc_depth <= minDepth
            zc = tc_depth[zcSel]

    #Horizontal weights
    uc_ptx, uc_weights = interp_weights(uc_coorddata,coordmesh)
    
    ###Array creation
    t = np.zeros( LpIter )                      #Time
    u = np.zeros((LpIter,coordmesh.shape[0]))   #U
    v = np.zeros((LpIter,coordmesh.shape[0]))   #V
    e = np.zeros((LpIter,coordmesh.shape[0]))   #eta

    ###Depth data
    b_file = Dataset(GridSpecFiPath, 'r')
    bdata = b_file.variables['depth_t']
    f = interp2d(b_file.variables['grid_x_T'][:], b_file.variables['grid_y_T'][:], bdata)
    bnew  = f(u_file.variables['xu_ocean'][:], u_file.variables['yu_ocean'][:])[uc_mask]

    ###Loop for interpolation to SLIM open boundary
    LpIter = 0
    LpIter2 = 0
    for j in range( len(EBran_files_list)):
        ###Import files
        e_file = Dataset(BranBase + 'eta/' + EBran_files_list[j], 'r')
        u_file = Dataset(BranBase + 'u/' + UBran_files_list[j], 'r')
        v_file = Dataset(BranBase + 'v/' + VBran_files_list[j], 'r')
        
        #Vertical weights. Part 2.         
        vweights = np.hstack([(zc[1]-zc[0])/2,(zc[2:]-zc[:-2])/2,(zc[-1]-zc[-2])/2])
        totalvw = np.zeros_like(u_file.variables['u'][0,zcSel,:,:])

        #Time        
        NoObs = e_file.variables['Time'].shape[0]
                                                            #Convert from days to seconds
        t[LpIter:LpIter+NoObs] = e_file.variables['Time'][:]*86400 + calendar.timegm((1979,1,1,0,0,0))         
        LpIter = LpIter + NoObs 
    
        #Time loop
        for k in range(NoObs):            
            uw = u_file.variables['u'][k,zcSel,:,:]*vweights[:,None,None]
            vw = v_file.variables['v'][k,zcSel,:,:]*vweights[:,None,None]
            totalvw[:] = vweights[:,None,None]
            totalvw[uw.mask] = 0
            vweight = np.sum(totalvw, axis=0)
            udata = (np.ma.sum(uw,axis=0)/vweight)[uc_mask]
            vdata = (np.ma.sum(vw,axis=0)/vweight)[uc_mask]
            
            #Interpolate t-cell eta to u-cell to match UV data
            edata = e_file.variables['eta_t'][k,:,:]
            f2 = interp2d(e_file.variables['xt_ocean'][:], e_file.variables['yt_ocean'][:], edata)
            enew  = f2(u_file.variables['xu_ocean'][:], u_file.variables['yu_ocean'][:])[uc_mask]
            
            #Depth plus tidal anomally
            Hdata = enew + bnew
            
            #Interpolate to slim
            ull = interpolate(udata*Hdata, uc_ptx, uc_weights)
            vll = interpolate(vdata*Hdata, uc_ptx, uc_weights)
            u[LpIter2,:], v[LpIter2,:] = lonlat_sys.rotate(ull,vll)
            e[LpIter2,:] = interpolate(enew,uc_ptx, uc_weights)
            LpIter2 = LpIter2 + 1
        
    ###Interpolate BRAN data in time. Add to Topex.
    t = t - 36000 #Readjust time to match input file
    print('Initial time')
    print( datetime.fromtimestamp(t[0]).strftime('%Y-%m-%d %H:%M:%S'))
    print('Final time')
    print( datetime.fromtimestamp(t[-1]).strftime('%Y-%m-%d %H:%M:%S'))

    #For tpxo 
    InitTime = datetime.fromtimestamp(t[0]-36000).strftime('%Y-%m-%d %H:%M:%S')
    FiTime   = datetime.fromtimestamp(t[-1]-36000).strftime('%Y-%m-%d %H:%M:%S')
    
    tslim = slimPre.Time(initial_time = InitTime, final_time = FiTime, time_step = TStep)
    osse,ou,ov = slimPre.tpxo_tide(region_open, tslim)
        
    print(t[0] - tslim._time[0])
    print(t[-1] - tslim._time[-1])

    ou   += interp1d(t,u,axis=0)(tslim._time)
    ov   += interp1d(t,v,axis=0)(tslim._time)
    osse += interp1d(t,e,axis=0)(tslim._time)
    
    ###Write...
    slimPre.write_file(data_dir + 'open_bnd.nc', region_open, tslim, [('eta',osse),('u',ou),('v',ov)])
    
#%%
prepro_base()
prepro_wind()
prepro_open()

###Exit
slimPre.exit(0)
