
""" script to perform particle simulations with behaviours of seagrass_out
writen October 2018. 
Edited July 2019.
Edited February 2020.

particle behaviours:
1)	Seeding from file of xy locations
2)	Mortality - Removed 
3)	Beaching to the coast with resuspension - Removed
4)	Wind drift
5)	Settling to the bottom over habitat - Modified to settle with decay function
6)	Keeping track of the seeding location
7)	Remove but count particles that leave the domain
    
New additions (Feb 2020):
    -File paths all moved to the start of the file
    -Change to seed from all mesh cells on seagrass (seeded from select
    cells in the sensitivity analysis)
    -Add loop to seed on selected dates in season with 1 execution
    -Modify code so seagrass can be near the bottom (velocity half of the
    water current)
    -Export start position in lat long
"""

################### imports #################

import slimParticle
import slimPre
import numpy as np
from osgeo import ogr

import os

#################### set up parameters/coefficent values ###################
#Start and end time for dispersal run
stTime, edTime = "2011-12-10 14:00:00", "2012-02-04 14:00:00"

###Tracker
dt = 300      #Time step between each iteration of the particle tracker
              #In seconds
        
#Start date/time of hydro simulation
init_time = '2011-08-28 14:00:00'

hydro_time = 1800   #Time step between each iteration of the hydrodynamic model

#Number of particles
ttl_n = 100008  #per seed location 
hr_n = np.round(ttl_n/24).astype(int) 


#Decay function
k = 0.075 #decay rate. per day.


#set up wind drag cooefficent:
#Indicate if particles are near the bottom. Y = yes. N = no(most runs).
NearBtm = 'N'
#The proportion of current that near bottom particles are affected by
btm = 0.5   

#% of wind velocity added to water velocity  for transport.
#Required if NearBtm = 'N'
percW = 1
wind_drift = percW/100

   
#Beaching distance
bd = 100.

#################### paths ####################
pathin = '/home/2/jc231398/TorresStrait/'
#pathin =  '/mnt/c/Users/schla/Desktop/TorresStraitModelling/'

#Ensure you change based on run parameters
years = '2011_2012'
mnth = 'Dec'
RunID = 'Spr_k_1WD'

#Set output path
pathout =  pathin +'PostPro/SensitivityAnalysis/' +mnth + '/' +RunID +'/' 
os.makedirs(pathout, exist_ok=True) 
                                                  
###Set file paths
MeshPath =  pathin + '/Mesh/torresV2.msh'
projection_mesh = '+proj=utm +zone=54 +south +ellps=WGS84'

bathPath = pathin +'Prepro/' +years +'_1/0/bathymetry.nc'
hydroPath = pathin + 'Hydro/' +years +'/'
windPath = pathin +'Prepro/' +years +'_1/0/wind.nc'

#################### functions ######################
def points_in_polygons(points,coordinate_system_of_points,polygonshp) :
    ds = ogr.Open(polygonshp)
    lyr_in = ds.GetLayer(0)
    point_ref=ogr.osr.SpatialReference()
    point_ref.ImportFromProj4(coordinate_system_of_points)
    ctran=ogr.osr.CoordinateTransformation(point_ref,lyr_in.GetSpatialRef())
    def check(x,y):
        [lon,lat,z]=ctran.TransformPoint(x,y)
        pt = ogr.Geometry(ogr.wkbPoint)
        pt.SetPoint_2D(0, lon,lat)
        lyr_in.SetSpatialFilter(pt)
        return len(lyr_in) > 0
    return np.array([check(points[i,0],points[i,1]) for i in range(points.shape[0])])

###################### Physical data ###############################
#set up data files and projections of the data
mesh =  slimPre.Mesh(MeshPath, projection_mesh)

#create the map for the difusivity based on the mesh size
slimPre.create_okubo_map(mesh, pathout +'okubo_map', 2e-4)

#get the distance to the coast of each node of the mesh for beaching function
region_global = slimPre.Region(mesh)
region_coast = slimPre.Region(mesh,'coast')
dist = slimPre.distance_msh(region_coast,50).evaluate_region(region_global)
slimPre.write_file(pathout +'distance_coast.nc', region_global,data=[('dist',dist)])


#get bathymetry
bathf = (bathPath,'bathymetry')
#read in the hydro data for the run
#mesh file, hydro directory, start of hydro files, time step
hydro = slimParticle.HydroReader(mesh, hydroPath, bathf, init_time, hydro_time)

#Loader
loader_wind = slimParticle.PreLoader(mesh,[(windPath,'u'),(windPath,'v')])
loader_kappa = slimParticle.PreLoader(mesh,[pathout +'okubo_map.msh'])
loader_dist = slimParticle.PreLoader(mesh,[(pathout +'distance_coast.nc','dist')])

#################### Log files ###################
#set up logFile for mortality function
logd_name= pathout +'decay.txt'
logd = open(logd_name,"w")
logd.close()

#set up logFile for settled [decayed and beached] particles
logf_name = pathout +'settled.txt'
logf = open(logf_name,'w')
logf.close()

#set up logFile for particles lost from open boundary
logob_name = pathout +'boundary.txt'
logob = open(logob_name,'w')
logob.close()

#set up logFile for particles lost through time
logl_name = pathout +'lost.txt'
logl = open(logl_name,'w')
logl.close()

#################### Seeding ###################
#set up an empty tracker for the particles
tracker = slimParticle.Tracker(mesh,["origin", "sse"])

#Set up the open boundaries so the particles can be removed
tracker.set_boundary_open(["south_west", "west", "north_east", "coral_sea", "south_east"])

#get the centre of the elements
c = mesh.compute_element_centroids()

#Sensitivity analysis. To seed from selected cells:
SelCells = np.array([503, 13325, 15726, 16160, 20316, 28192, 
                     31798, 37368, 41527, 41698])
pos = c[SelCells,:2]
tags = np.array([ SelCells, SelCells ]).T
  
#print("add the particles to the tracker: location, tag, repeats, ignore those outside?")
added = tracker.add_particles(pos, tags, hr_n, True)
C0 = hr_n*pos.shape[0]

##################### Dipersal #############################
#Range-Kutta - start time, end time, dt
#Set ten hours before. Coordinated Universal Time (UTC) to Australian Eastern Standard Time.
rk = slimParticle.ERK44(stTime, edTime, dt)

#Get initial time
t0 = rk.time

iout = 0
print("begining loop  " +mnth +'  ' +RunID)

##Decay function
#And Set concentration at t=n-1=0-1 to concentration at t=n=0
Cnm1 = C0

#start the time loop for the simulation
while not rk.end:
    #Time
    tn = (rk.time-t0)/(24*60*60) #Time in days at step n
    
    #Runge-Kutta sub time steps to displace particles
    for irk in range(rk.nsub) :
        hydro.load(tracker,rk.time)
        #If particles are not near the bottom, move them with 
        #the water velocity and the wind velocity
        if NearBtm == 'N':
            #Load wind
            wind = loader_wind.load(tracker,rk.time)   
            #move the particles including the wind drift coefficient
            dx = (hydro.uv + wind*wind_drift)*rk.dt
        
        #If the particles are near the bottom, move them with only 
        #half (btm = 0.5) of the water velocity.
        else:
            dx = hydro.uv*btm*rk.dt
        
        #Sub time step
        rk.sub_time_step(tracker,dx)
            
    #diffusivity
    hydro.load(tracker,rk.time) #required for diffusivity
    kappa = loader_kappa.load(tracker,rk.time)
    rk.end_time_step(tracker, hydro.diffusivity(kappa,rk.dt))

    ###Settlement and loss
    #________________________________________________________________________
    #settling - decay        
    if np.logical_and(tn < 23/24, abs(rk.time%3600) < 1.):
        ###Seeding in firt 24 hours
        added = tracker.add_particles(pos, tags, hr_n, True)
        C0 = C0 + hr_n*pos.shape[0]

        tnm1 = (rk.time-rk.dt-t0)/(24*60*60) #Time in days at step n-1
        Cnm1 = C0*np.exp(-k*tnm1) #Decay function t n-1 
        
        Cn = C0*np.exp(-k*tn) #Decay function tn.

        #Calculate the difference in concentration at n-1 vs n
        Chng = Cnm1 - Cn
        #########
        
    else: 
        Cn = C0*np.exp(-k*tn) #Decay function tn.
        Chng = Cnm1 - Cn
        Cnm1 = Cn  #t n-1
 
    #Boolean array to select decayed particles 0 = False, not decayed 
    Bdecay = np.zeros(tracker.n_particles, dtype = bool)
    
    #Apply prop. to new base(i.e. number of particles), given additional sources of loss
    TakeB = np.round( (Chng/C0)*tracker.n_particles).astype(int)
    
    if TakeB >= 1:       
    #                           Number of particles, number decayed
        decay = np.random.choice(tracker.n_particles, TakeB, 
                             replace = False)
        Bdecay[decay] = True   #Set index of decayed particles to true
    
    #________________________________________________________________________
    #Write mortality to file 
    logd = open(logd_name,"ab")  #Mode - a = append, b = binary
    decPrint = np.array( [[Cn, TakeB, tn]])
    np.savetxt(logd, decPrint, delimiter=',', newline='\n') 
    logd.close()
    
    #________________________________________________________________________
    #settling - beaching
    
    distance = loader_dist.load(tracker, rk.time)
    on_beach = distance < bd  
    
    #________________________________________________________________________
    #Get all settled particles
    ttl_set = np.logical_or(Bdecay, on_beach[:,0])
    
    #Save settled particles to file for analysis
    logf = open(logf_name,"ab")  #Mode - a = append, b = binary
    np.savetxt(logf, np.concatenate((tracker.position[ttl_set], tracker.data[ttl_set][:,1:],
                                    np.ones(tracker.data[ttl_set][:,1:].shape)*tn), axis = 1), 
                                    delimiter=',', newline='\n')  #Decay
    logf.close()
    
    #________________________________________________________________________
    #remove the settled and lost particles from the simulation
    #identify particles on an open boundary to be lost
    on_boundary = tracker.is_on_open_boundary()
    
    #Count particles lost from open boundary. Save to file for analysis
    logob = open(logob_name,"ab")  #Mode - a = append, b = binary
    
    ###edit as per above###
    np.savetxt(logob, np.concatenate((tracker.position[on_boundary], tracker.data[on_boundary][:,1:],
                                    np.ones(tracker.data[on_boundary][:,1:].shape)*tn), axis = 1), 
                                    delimiter=',', newline='\n')  #boundary
    logob.close()
    
    remove = np.logical_or(ttl_set, on_boundary)
    tracker.remove_particles(remove)
    #this changes the length of the tracker so be careful!
    
    #________________________________________________________________________
    #Track loss through time
    logl = open(logl_name,"ab")  #Mode - a = append, b = binary
    loss = np.array( [[np.sum(Bdecay), np.sum(on_beach[:,0]), np.sum(on_boundary), tn]])
    np.savetxt(logl, loss, delimiter=',', newline='\n') 
    logl.close()
    
    #_______________________________________________________________________
    #Track loop interations
    if abs(rk.time%3600) < 1.:
        print(iout)
        iout += 1
