# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 13:50:52 2019

@author: schla

Conduct sensitivity analysis on Torres Strait seagrass dispersion
following Simons et al. (2013):
    
Simons RD, Siegel DA, Brown KS (2013) 
Model sensitivity and robustness in the estimation of larval transport: 
A study of particle tracking parameters. 
J Mar Syst 119:19-29. 
doi:10.1016/j.jmarsys.2013.03.004

Modified 08/01/2020
Changed to conduct analysis on SLIM variable resolution mesh,
rather than imposing a square mesh. Changed due to senditivity of analysis
to the percentage area (i.e. size) of the squares in the imposed mesh.

"""
#%%
###Module imports
import numpy as np

import geopandas as gpd

from scipy.spatial import cKDTree
from scipy.stats import rv_histogram 

import os
import time

import slimPre

#%%
###Parameters etc.

#Folder with run data.
mnth = 'OctT2'
RunDir = 'btwn_Dk_2WD'

#File path in
pathin = '/home/2/jc231398/TorresStrait/'
wrkdir = pathin +'PostPro/SensitivityAnalysis/' +mnth +'/' +RunDir

#Number of nearest neighbours to search to assign
#Mesh cell to settled particles
NoNN = 7

#Focal point weight
Wgt = 1

#Number of smoothing iterations
SmthIter = 1

#Fill value
fill = -99

#Seed cells
#SelCells = np.array([503, 13325, 15726, 16160, 20316, 28192, 
#                     31798, 37368, 41527, 41698])
SelCells = np.array([31798, 37368, 41527, 41698])

###Generate randoms no's. of particles to randomly select
rn = np.array([10, 25, 50, 100, 175, 250, 
               500, 1000, 1500, 2500, 3500, 5000])

NoRnSel = 500 #Number of times rn no. of particles are to be randomly selected

#%%
###Mesh vertices and nearest neighbour for centroids

#Mesh file and projection
MeshFi = pathin + 'Mesh/torresV2.msh'
mesh_projection = "+proj=utm +zone=54 +south +ellps=WGS84"

#Mesh object
mesh = slimPre.Mesh(MeshFi, mesh_projection)

#Get mesh vertices
groups = mesh._groups
MVertXYZ = [groups.coordinatesDof().getGroupProxy(i) for i in range(groups.getNbElementGroups())]
MVert = np.asarray(MVertXYZ)[0,:,:2,:]  #Get rid of redundant dimension and z coordinates

#3x mesh nodes nearest neighbour kd-tree's
MVertNN_0 = cKDTree(MVert[0,:,:].T)
MVertNN_1 = cKDTree(MVert[1,:,:].T)
MVertNN_2 = cKDTree(MVert[2,:,:].T)

#get the centre of the elements. Remove z value
c = mesh.compute_element_centroids()[:,:2]

#For unique mesh cells
UniMesh = np.arange(0, c.shape[0], 1)

#Centroid nearest neighbour
CentroidNN = cKDTree(c)

#%%
#%%
#Functions

### In/out of triangle function following:
'''
Gamedev-Maths
https://github.com/SebLague/Gamedev-Maths/blob/master/PointInTriangle.cs
'''

def PointInTri(v0, v1, v2, p):
    numer = v0[0]*(v2[1]-v0[1]) + (p[1]-v0[1])*(v2[0]-v0[0]) - p[0]*(v2[1]-v0[1])
    denom = (v1[1]-v0[1])*(v2[0]-v0[0]) - (v1[0]-v0[0])*(v2[1]-v0[1])
    w1 = numer/denom
    
    w2 = (p[1] - v0[1] -w1*(v1[1]-v0[1]) )/(v2[1]-v0[1])
    
    #Test in triangle
    #w1,w2 >= 0 and w1+w2<=1
    inTest = np.logical_and(np.logical_and(w1>=0, w2>=0), w1+w2<=1)

    return inTest

#%%
### Area of a triangle:
#Shoelace formula for area of polygon
def PolyArea(x,y):
    return 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))
    
#%%
###Function to get ID's of triangles surrounding focal triangle
def SurroundingTri(MeshCellID):
    #Vertices of focal triangle
    v0 = MVert[0,:,MeshCellID]
    v1 = MVert[1,:,MeshCellID]
    v2 = MVert[2,:,MeshCellID]
        
    #Get indices of surrounding triangles:
    #v0 test. Return index if same location as v0. Across first, second, and third
    #listed vertices. 
    ind0 = np.hstack( (MVertNN_0.query_ball_point(v0, 0), MVertNN_1.query_ball_point(v0, 0),  
                       MVertNN_2.query_ball_point(v0, 0) ) )
        
    #v1 test
    ind1 = np.hstack( (MVertNN_0.query_ball_point(v1, 0), MVertNN_1.query_ball_point(v1, 0),  
                       MVertNN_2.query_ball_point(v1, 0) ) )
        
    #v2 test
    ind2 = np.hstack( (MVertNN_0.query_ball_point(v2, 0), MVertNN_1.query_ball_point(v2, 0),  
                       MVertNN_2.query_ball_point(v2, 0) ) )
        
    #Get unique surrounding triangle ID's
    STri = np.unique( np.hstack( (ind0, ind1, ind2) ) )

    #Remove focal triangle
    STri = STri[STri != MeshCellID]
    
    return STri

#%%
###Inverse distance weighted Interpolation    
'''
Modified from: 
https://stackoverflow.com/questions/3104781/inverse-distance-weighted-idw-interpolation-with-python

e.g.
zi = simple_idw(c[UniqInd,0], c[UniqInd,1], TmpConc, #:Interpolated from
                  c[UniqSetMesh[lb],0], c[UniqSetMesh[lb],1])   #:to

'''
def simple_idw(x, y, z, xi, yi):
    #Pythagoras theorem to get eucildean distance
    dist =  np.sqrt( (x-xi)**2 + (y-yi)**2 )

    # In IDW, weights are 1 / distance
    weights = 1.0 / dist

    # Make weights sum to one
    weights /= weights.sum(axis=0)

    # Multiply the weights for each interpolated point by all observed Z-values
    zi = np.dot(weights.T, z)
    return zi

#Alternate:
#from scipy.interpolate import Rbf
#Interpolation function. x,y,z of centroids 
#rbfi = Rbf(c[STri,0], c[STri,1], np.zeros(STri.shape[0]), TmpConc, 
#             function = 'linear')
        
#Apply interpolation function for focal point
#di = rbfi(c[UniqSetMesh[lb],0], c[UniqSetMesh[lb],1], 0)
    
#%%
###Weighted average procedure
'''
Inputs:
>TriID. Array of focal triangle ID's
>TriDict. Dictionary with surrounding triangles
>TriConc.List of concentration for focal triangles
'''

def WAProc(TriID, STriDict, TriConc):          
    #Make vector to store weighted average concentrations
    WAVec = np.zeros(TriID.shape)
        
    #Weighted average of triangles that share vertices              
    for lb in range(TriID.shape[0]):
        #Get ID's of surrounding triangles
        STri = STriDict[TriID[lb]]
            
        #Find if any STris have particles settled in them
        #i.e. also in TriID.
        #if yes, then save in temporary concentration array:
        TmpConc = np.zeros(STri.shape)
            
        for i in range(STri.shape[0]):
            if str(any(TriID == STri[i])) == 'True':
                TmpConc[i] = TriConc[TriID == STri[i]]
                                       
        di = simple_idw(c[STri,0], c[STri,1], TmpConc, #:Interpolated from
        c[TriID[lb],0], c[TriID[lb],1])   #:to
            
        denom = STri.shape[0]+Wgt
        #Average interpolated and true value
        WAVec[lb] = di*STri.shape[0]/denom + TriConc[lb]*Wgt/denom
        
    return WAVec
 
#%%
#%%

###Sensititvity analysis:
print(mnth, RunDir)

###Import Settlement locations and initial seed location tag.  
FiPath = wrkdir +'/settled.txt'

Settled = np.loadtxt(FiPath, delimiter = ",")
SetX, SetY = Settled[:,0], Settled [:,1]
Seed = Settled[:,2]

#set up logFile for rn @ 95% confidence
logC_name="_Conf95.txt"
logC_path = wrkdir +'/' +RunDir +logC_name
logC = open(logC_path,"w")
logC.close()
           
#set up logFile for FUVup
logF_name="_FUVup.txt"
logF_path = wrkdir +'/' +RunDir +logF_name
logF = open(logF_path,"w")
logF.close()

#Start loop
for eb in range(SelCells.shape[0]):
    
    ##Select settlement locations by initial seed
    SetX_Sd = SetX[Seed == SelCells[eb]]
    SetY_Sd = SetY[Seed == SelCells[eb]]
    
    ##Make arrays for: 
    #for unique particle id
    Unique = np.arange(np.sum(Seed == SelCells[eb]))
    
    #to assign 'settled' mesh cell
    SetMesh = np.full(np.sum(Seed == SelCells[eb]), fill)

#%%
    ###Loop through settled particles.
    ###For each particle, assign mesh cell that the particle settled in
    
    #Loop
    for i in range(Unique.shape[0]):
        XY = np.array([SetX_Sd[i], SetY_Sd[i] ])
    
        #Query mesh centroid nearest neighbour kd-tree
        
        #idp - index of 'nearest neighbour' 
        #from settled particle to mesh centroids
        dp,idp = CentroidNN.query(XY, k = NoNN)
        
        #Search if particle is in one of the 7(NoNN) triangles of the
        #7 nearest nearest neighbor centroids
        for j in range(idp.shape[0]):
            #Mesh triangle vertices of jth nearest neighbour
            v0 = MVert[0,:,idp[j]]
            v1 = MVert[1,:,idp[j]]
            v2 = MVert[2,:,idp[j]]
            
            if str( PointInTri(v0, v1, v2, XY)) == 'True':
                SetMesh[i] = UniMesh[idp[j]]
                #if j != 0:
                #    print(j)
                break
            else:
                continue
        
    #Assign mesh nearest neighbour for particles physically just outside model
    #Mesh cells (very few)
    if str(any(SetMesh == fill)) == 'True':
        XY = np.zeros( (np.sum(SetMesh == fill), 2) )
        XY[:,0] = SetX_Sd[SetMesh == fill] #x
        XY[:,1] = SetY_Sd[SetMesh == fill] #y
        
        #Get first nearest neighbor for all particles outside mesh
        dp,idp = CentroidNN.query(XY) 
        
        #Assign first nearest neighbor as mesh settled
        SetMesh[SetMesh == fill] = UniMesh[idp]
                        
    del(i,j)

#%%
    ###Count number of particles by unique cells for ***reference solution***
    
    #Get unique cells
    UniqSetMesh = np.unique(SetMesh)

    #Generate array for counts by mesh cell
    Count = np.zeros(UniqSetMesh.shape)
    
    #Generate array to store area:
    TriArea = np.zeros(UniqSetMesh.shape)
    
    #Generate concentration array for counts standardized by triangle
    #Area
    conc = np.zeros(UniqSetMesh.shape)
    
    for k in range(UniqSetMesh.shape[0]):
        #Count number of particles in cells
        Count[k] = np.sum(SetMesh == UniqSetMesh[k])
        
        #Get unique mesh vertices
        v0 = MVert[0,:,UniqSetMesh[k]]
        v1 = MVert[1,:,UniqSetMesh[k]]
        v2 = MVert[2,:,UniqSetMesh[k]]
        
        #Get Area of Triangle
        TriArea[k] = PolyArea([v0[0], v1[0], v2[0]],[v0[1], v1[1], v2[1]])
        
        #Divide count by area to get #Settled per m2
        conc[k] = Count[k]/TriArea[k]
    
    del(k)

    #Perform Simons et al. PDD step 2. 
    #Divide each #settled per m2 by the 
    #sum of the sum of the # settled per m2 to
    #convert to distribution
    DivConc = conc/np.sum(conc)
    
    #Check. For each seed location: Number that settled should equal number counted in cells
    #Extra check if it is a probability distribution: sum(DivConc) = 1
    print(SelCells[eb], np.sum(Seed == SelCells[eb]), np.sum(Count), sum(DivConc))
  
#%% #Simons PDD step 3. 'Gaussian filter' analogue part 1     
    #Build surrounding triangle dictionary for
    #All triangles in 'filter' area i.e. 2 layers of 
    #triangles around triangles with settled particles
    FiltID = UniqSetMesh
    layer  = np.ones(UniqSetMesh.shape)
    
    #Make blank dictionary to store surrounding triangles
    STriDict = {}
    
    #Get ID's of surrounding triangles for 'first layer'
    for st in range(UniqSetMesh.shape[0]):
        STri = SurroundingTri(UniqSetMesh[st]).astype(int)
        
        #Add to dictionary
        STriDict[UniqSetMesh[st]] = STri 
                   
        for i in range(STri.shape[0]):
            #Identify unrecorded surrounding triangle in 'second layer'
            #i.e. not in FiltID_l2
                
            if str(any(FiltID == STri[i])) == 'False':
                #Add to FiltID
                FiltID = np.concatenate( (FiltID, [STri[i]]), axis = 0 )
                layer  = np.concatenate( (layer,  [2]), axis = 0)
                    
                #Calculate surrounding triangles and add to dictionary
                STri2 = SurroundingTri(STri[i]).astype(int)
                STriDict[STri[i]] = STri2
      
        
    #Get ID's of surrounding triangles for 'second layer'
    for st2 in range(FiltID.shape[0]):
        #Calc surrounding tri's for tri's in first and second layer
        STri2 = SurroundingTri(FiltID[st2]).astype(int)
            
        for j in range(STri2.shape[0]):                
            #Identify unrecorded surrounding triangle in 'third layer'
            #i.e. not in FiltID_l2
                
            if str(any(FiltID == STri2[j])) == 'False':
                #Add to FiltID
                FiltID = np.concatenate( (FiltID, [STri2[j]]), axis = 0 )
                layer  = np.concatenate( (layer,  [3]), axis = 0)
                
                #Calculate surrounding triangles and add to dictionary
                STri3 = SurroundingTri(STri2[j]).astype(int)
                STriDict[STri2[j]] = STri3
    
#%%
    #Make a concentration vector for layers 1 and 2
    TriID_l1l2 = FiltID[np.logical_or(layer ==1, layer==2)]
    Conc_l1l2 = np.zeros( TriID_l1l2.shape )
    
    for k in range(UniqSetMesh.shape[0]):
        if str(any(TriID_l1l2 == UniqSetMesh[k])) == 'True':
            Conc_l1l2[TriID_l1l2 == UniqSetMesh[k]] = DivConc[k]
                
#%%
    #Simons PDD step 3. 'Gaussian filter' analogue part 2
    #Inverse distance weighted Interpolations 
    
    #Layer 1 and layer 2 IDW interpolation      
    WAConc_l1l2 = WAProc(TriID_l1l2, STriDict, Conc_l1l2)
    
    
    #Make a concentration vector to go with FiltID
    FiltConc = np.zeros( FiltID.shape)
    
    for k in range(TriID_l1l2.shape[0]):
        if str(any(FiltID == TriID_l1l2[k])) == 'True':
            FiltConc[FiltID == TriID_l1l2[k]] = WAConc_l1l2[k]
    
    #Perform IDW interpolation with all three layers
    WAConc = WAProc(FiltID, STriDict, FiltConc)
    
    #Smooth
    for lp in range(SmthIter):
        if lp == 0:
            SmthArray = np.zeros( (SmthIter, FiltID.shape[0]) )
            SmthArray[lp,:] = WAProc(FiltID, STriDict, WAConc)
        else:
            SmthArray[lp,:] = WAProc(FiltID, STriDict, SmthArray[lp-1,:])
    RefConc = SmthArray[-1,:]
    
#%%
    ###Subsampling 
    
    ##Check number of settled particles from seed loc
    ## > numbers in rn
    rnSub = rn[ np.sum(Seed == SelCells[eb]) > rn ]

    ##Perform random selection of rn particles NoRnSel times
    #Calculate and store FUV for selected particles
    
    ###Generate array to put FUV from each iterations
    FUVar = np.empty( (rnSub.shape[0], NoRnSel), dtype = float)         
    
    #Loop
    for a in range(FUVar.shape[0]):
        for b in range(FUVar.shape[1]):
            print(rn[a], b, time.process_time())
            
            #Randomly select particles
            RSel = np.random.choice(Unique, rn[a], replace = False)
    
            SubSetMesh = SetMesh[RSel]
            
            TmpCount = np.zeros(UniqSetMesh.shape) #Make temporary count Array
            TmpConc  = np.zeros(UniqSetMesh.shape) #Make temporary concentration Array
            
            #Count how many particles in grids from subsample
            for i in range(UniqSetMesh.shape[0]):
                TmpCount[i] = np.sum(SubSetMesh == UniqSetMesh[i])
                
                #Divide count by area to get #Settled per m2
                TmpConc[i] = TmpCount[i]/TriArea[UniqSetMesh == UniqSetMesh[i]]
            
            #Simons step 2
            DivTmp = TmpConc/sum(TmpConc)
                        
#%%         Simons PDD step 3. 'Gaussian filter' analogue part 2
            #Inverse distance weighted Interpolations 
            
            #Make a concentration vector for layers 1 and 2
            Tmp_l1l2 = np.zeros( TriID_l1l2.shape )
            
            for k in range(UniqSetMesh.shape[0]):
                if str(any(TriID_l1l2 == UniqSetMesh[k])) == 'True':
                    Tmp_l1l2[TriID_l1l2 == UniqSetMesh[k]] = DivTmp[k]
                        
            #Layer 1 and layer 2 IDW interpolation      
            WATmp_l1l2 = WAProc(TriID_l1l2, STriDict, Tmp_l1l2)
            
            #Make a concentration vector to go with FiltID
            FiltTmp = np.zeros( FiltID.shape)
            
            for k in range(TriID_l1l2.shape[0]):
                if str(any(FiltID == TriID_l1l2[k])) == 'True':
                    FiltTmp[FiltID == TriID_l1l2[k]] = WATmp_l1l2[k]
            
            #Perform IDW interpolation with all three layers
            WATmp = WAProc(FiltID, STriDict, FiltTmp)
            
            #Smooth
            for lp in range(SmthIter):
                if lp == 0:
                    SmthArray = np.zeros( (SmthIter, FiltID.shape[0]) )
                    SmthArray[lp,:] = WAProc(FiltID, STriDict, WATmp)
                else:
                    SmthArray[lp,:] = WAProc(FiltID, STriDict, SmthArray[lp-1,:])
            SubConc = SmthArray[-1,:]
            
#%%
            #Calculate FUV. Save in FUV array. Row = Size of sample. 
            #                                  Column = Repeat sampling, NoRnSel.
            FUVar[a,b] = 1 - (np.corrcoef(RefConc, SubConc)[1,0])**2
            
            del(RSel, TmpConc)
            
#%%
    #Calculate FUV upper bound
    FUVup = np.zeros(rnSub.shape[0])

    for hc in range(FUVup.shape[0]):
        FUVhist = np.histogram(FUVar[hc,:], bins = 'auto')
        SCIhist = rv_histogram(FUVhist)
        FUVup[hc] = SCIhist.ppf(0.95)
    
    if rnSub.shape[0] < rn.shape[0]:
        FillVec = np.ones ( rn.shape[0] - rnSub.shape[0] )*fill
        FUVup = np.concatenate( (FUVup, FillVec), axis = None)

    #Save FUV upper limit
    FUVsv = np.concatenate( ([SelCells[eb]], FUVup), axis = None)
    logF = open(logF_path,"ab")  #Mode - a = append, b = binary
    np.savetxt(logF, np.array([FUVsv]), 
               delimiter=',', newline='\n')
    logF.close()
           
#%%
    #Calculate intercept of FUV upper bound with FUV == 0.05. 
    #Use equation of line.
    
    if np.logical_and(np.any(FUVup >= 0.05), np.any(FUVup < 0.05)): 
        #Get x(rn), y(FUV) for greater the 0.05
        FUVgreater = FUVup[FUVup >= 0.05][-1]
        RNgreater = rn[FUVup >= 0.05][-1]
        
        #Get x(rn), y(FUV) for lesser. One ind after greater
        lesserLog = (FUVup == FUVgreater)
        LsLogFi = np.zeros( lesserLog.shape[0], dtype = bool)
        
        for lb in range(lesserLog.shape[0]):
            if str(lesserLog[lb]) == 'True':
                LsLogFi[lb+1] = True
                  
        FUVlesser = FUVup[LsLogFi]            
        RNlesser = rn[LsLogFi]      
        
        #Calculate intercept of line connecting 'greater' and 'lesser' with FUV = 0.05
        grad = (FUVgreater - FUVlesser)/ (RNgreater - RNlesser)
        yint = FUVgreater - grad*(RNgreater)
        
        intCI = (0.05 - yint)/grad 
        
        #Export intCI
        logC = open(logC_path,"ab")  #Mode - a = append, b = binary
        np.savetxt(logC, np.array([[SelCells[eb], UniqSetMesh.shape[0], 
                        sum(TriArea), intCI]]), delimiter=',', newline='\n')
        logC.close()
        
    else:
        print('Out of range')
        
        #Add fill value to concentration file
        logC = open(logC_path,"ab")  #Mode - a = append, b = binary
        np.savetxt(logC, np.array([[SelCells[eb], UniqSetMesh.shape[0], 
                        sum(TriArea), fill]]), delimiter=',', newline='\n')
        logC.close()
        
    #%%
       
