# -*- coding: utf-8 -*-
"""
Created on Fri May  1 13:25:48 2020

@author: schla

Generate a connectivity matrix of all 7000+ seed locations
for SLIM Torres Strait model.

"""

#%%
###Import packages
import numpy as np

import geopandas as gpd

from scipy.spatial import cKDTree
from scipy.stats import rv_histogram 

import os
import re
import time

import slimPre

#%%
###Parameters etc.

#Folder with run data.
years = '2011_2012'
Np = '100'
st, ed = 9, 17  #25=0:8;    100=9:17;   2500=18:26.
            
#File path in
pathin = '/home/2/jc231398/TorresStrait/'
run_dir = pathin +'PostPro/SensitivityAnalysis/Phase2/R2/' +years

#%%
###Mesh vertices and nearest neighbour for centroids

#Mesh file and projection
MeshFi = pathin + 'Mesh/torresV2.msh'
mesh_projection = "+proj=utm +zone=54 +south +ellps=WGS84"

#Mesh object
mesh = slimPre.Mesh(MeshFi, mesh_projection)

#Get mesh vertices
groups = mesh._groups
MVertXYZ = [groups.coordinatesDof().getGroupProxy(i) for i in range(groups.getNbElementGroups())]
MVert = np.asarray(MVertXYZ)[0,:,:2,:]  #Get rid of redundant dimension and z coordinates

#get the centre of the elements. Remove z value
c = mesh.compute_element_centroids()[:,:2]

#Centroid nearest neighbour
CentroidNN = cKDTree(c)

#For unique mesh cells
UniMesh = np.arange(0, c.shape[0], 1)

#%%
###Seed triangles
STriPath = pathin +'PostPro/SensitivityAnalysis/Phase2/' +'NodeElements.csv'
SeedTris = np.loadtxt(STriPath, delimiter = ",", skiprows =1)[:,0].astype(int)

#For unique seed triangles. For matrix rows and columns.
UniSeedMesh = np.arange(0, SeedTris.shape[0], 1)

#%%
#%%
#Functions

### In/out of triangle function following:
'''
Gamedev-Maths
https://github.com/SebLague/Gamedev-Maths/blob/master/PointInTriangle.cs
'''

def PointInTri(v0, v1, v2, p):
    numer = v0[0]*(v2[1]-v0[1]) + (p[1]-v0[1])*(v2[0]-v0[0]) - p[0]*(v2[1]-v0[1])
    denom = (v1[1]-v0[1])*(v2[0]-v0[0]) - (v1[0]-v0[0])*(v2[1]-v0[1])
    w1 = numer/denom
    
    w2 = (p[1] - v0[1] -w1*(v1[1]-v0[1]) )/(v2[1]-v0[1])
    
    #Test in triangle
    #w1,w2 >= 0 and w1+w2<=1
    inTest = np.logical_and(np.logical_and(w1>=0, w2>=0), w1+w2<=1)

    return inTest


#%%

### Find the triangle that a particle settled in

#Fill value
fill = -99

#Number of nearest neighbours to search to assign
#Mesh cell to settled particles
NoNN = 7

def FindTri(X, Y):
    SetTri = fill
    XY = np.array([X,Y])
    
    #Query kd-tree
    #idp - index of 'nearest neighbour' 
    #from particle to mesh centroids
    dp,idp = CentroidNN.query( XY, k = NoNN)
    
    #Search if particle is in one of the 7(NoNN) triangles of the
    #7 nearest nearest neighbor centroids
    for lb in range(idp.shape[0]):
        #Mesh triangle vertices of jth nearest neighbour
        v0 = MVert[0,:,idp[lb]]
        v1 = MVert[1,:,idp[lb]]
        v2 = MVert[2,:,idp[lb]]
        
        if str( PointInTri(v0, v1, v2, XY)) == 'True':
            SetTri = UniMesh[idp[lb]]
            break
        else:
            continue
    
    #Assign mesh nearest neighbour for particles physically just outside model
    #Mesh cells (very few)
    if str(SetTri == fill) == 'True':
        #Assign first nearest neighbor as mesh settled
        SetTri = UniMesh[idp[0]]
        
    return SetTri
            
#%%
#%% 
###Extract data from files. Parse in loop. 
#Dates
dates = next(os.walk(run_dir +'/'))[1]        
dates.sort()    


#Get run directories
#runs
runs = next(os.walk(run_dir +'/' +dates[0]))[1]
runs.sort()

#Select folders by no. of particles seeded
rSel = np.zeros(len(runs), dtype = bool)

for lb in range(len(runs)):
    DirEnd = runs[lb][0:2]    #Get start of dir name

    if np.logical_and( int(DirEnd) >= st, int(DirEnd) <= ed):
         rSel[lb] = 1 
     
runsFinal = np.array(runs)[rSel]

#set up names for saved results files
logTri_name = "_Triangles.csv"
logOH_name  = "_OnHabitat.csv"

#Generate empty connectivity matrix to fill.
ConMat = np.zeros( (SeedTris.shape[0], SeedTris.shape[0]) )

#Count particles in node for con mat. In loop.
#Loop dates
for i in range(len(dates)):
    #Loop runs
    for j in range(runsFinal.shape[0]):
        print(dates[i], runsFinal[j])
        
        wrkdir = run_dir +'/' +dates[i] +'/' +runsFinal[j] +'/'
        
        ###Import seed and settlement locations 
        FiPath = wrkdir +'/settled.txt'
        
        Settled = np.loadtxt(FiPath, delimiter = ",")
        edX, edY = Settled[:,0], Settled [:,1]
        stX, stY = Settled[:,2], Settled [:,3]
        SeedTriImp = Settled[:,5]        

        #Boolean array to distinquish particles that settle on habitat. 
        SetOnHab = np.zeros(edX.shape[0], dtype = bool)
        
        #to assign 'settled' mesh cell
        SetTriAll = np.full(edX.shape[0], fill)
        
        #Loop through particles. For each, find triangle they settled in.  
        for k in range(edX.shape[0]):
            SetTri = FindTri(edX[k], edY[k])
            SetTriAll[k] = SetTri
            
            #Change SetOnHab if particle settled on habitat
            if any(SeedTris == SetTri):
                SetOnHab[k] = 1
                
                #Get row and column indices for connectivity matrix
                row = UniSeedMesh[SeedTris == SetTri]
                col = UniSeedMesh[SeedTris == SeedTriImp[k]]
                print(SeedTriImp[k], col)
                #Fill connectivity matrix
                ConMat[row, col] = ConMat[row, col] +1
                

        #Save triangles
        logTri_path = wrkdir +'/' +runsFinal[j] +logTri_name
        np.savetxt(logTri_path, np.array([SeedTriImp[SetOnHab], SetTriAll[SetOnHab]]).T,
                   delimiter = ',')
        
        #Save ttl no. vs no. on habitat
        logOH_path  = wrkdir +'/' +runsFinal[j] +logOH_name
        np.savetxt(logOH_path, np.array([SetOnHab.shape[0], np.sum(SetOnHab)]).T,
                   delimiter = ',')


#Save connectivity matrix
np.savetxt(run_dir +'/' +'ConMatR1_' +Np +'.csv', ConMat, delimiter = ',')

